@@ -55,6 +55,12 @@
 #include "map.h"
 #include "back-sch.h"
 
+/*
+ * This is not a public function in 389-ds but we need to use it in backend_search_cb(),
+ * see a comment there on the difference between slapi_send_ldap_result() and send_ldap_result().
+ */
+extern void send_ldap_result(Slapi_PBlock *pb, int err, char *matched, char *text, int nentries, struct berval **urls);
+
 static int backend_passwdmod_extop(Slapi_PBlock *pb);
 backend_extop_handlers_t extop_handlers[] = {{EXTOP_PASSWD_OID, (IFP) backend_passwdmod_extop},
 					    {NULL, NULL}};
@@ -432,13 +438,8 @@ backend_set_operational_attributes(Slapi_Entry *e,
 	 * plugin want to override the values using the configuration, they
 	 * can. */
 	if (gmtime_r(&timestamp, &timestamp_tm) == &timestamp_tm) {
-		sprintf(timestamp_str, "%04d%02d%02d%02d%02d%02dZ",
-			timestamp_tm.tm_year + 1900,
-			timestamp_tm.tm_mon + 1,
-			timestamp_tm.tm_mday,
-			timestamp_tm.tm_hour,
-			timestamp_tm.tm_min,
-			timestamp_tm.tm_sec);
+		strftime(timestamp_str, sizeof(timestamp_str),
+			 "%Y%m%d%H%M%SZ", &timestamp_tm);
 		slapi_entry_add_string(e, "createTimestamp", timestamp_str);
 		slapi_entry_add_string(e, "modifyTimestamp", timestamp_str);
 	}
@@ -1524,16 +1525,17 @@ backend_entries_to_return_push(struct backend_search_cbdata *cbdata, Slapi_Entry
 	struct entries_to_send *e_to_send = NULL;
 	struct cached_entry *entry = NULL;
 	bool_t dont_cache = FALSE;
+	void *ndn_key = (void *) slapi_entry_get_ndn(e);
 	PLHashTable* ht = (PLHashTable*) cbdata->state->cached_entries;
 
-	if ((cbdata == NULL) || (e == NULL)) return;
+	if ((cbdata == NULL) || (e == NULL) || (ndn_key == NULL)) return;
 
 	e_to_send = (struct entries_to_send *) slapi_ch_calloc(1, sizeof(struct entries_to_send));
 
 	dont_cache = cbdata->state->use_entry_cache ? FALSE : TRUE;
 
 	if (!wrap_rwlock_wrlock(cbdata->state->cached_entries_lock)) {
-		entry = PL_HashTableLookup(ht, slapi_entry_get_ndn(e));
+		entry = PL_HashTableLookup(ht, ndn_key);
 		if (entry != NULL) {
 			/* There is an entry in the hash table but is it the same? */
 			char *e_modifyTimestamp = slapi_entry_attr_get_charptr(e, "modifyTimestamp");
@@ -1647,6 +1649,45 @@ backend_send_mapped_entries(struct backend_search_cbdata *cbdata)
 	cbdata->entries_tail = NULL;
 }
 
+static void
+backend_search_free_staged(struct backend_search_cbdata *cbdata)
+{
+	struct backend_staged_search *staged, *next;
+	int i;
+	int nb_entries = 0;
+	int nb_stages = 0;
+
+	if (cbdata == NULL) {
+		return;
+	}
+
+	staged = cbdata->staged;
+	while (staged != NULL) {
+		nb_stages++;
+		for (i = 0;
+				(i < staged->count) &&
+				(staged->entries != NULL) &&
+				(staged->entries[i] != NULL);
+				i++) {
+			slapi_entry_free(staged->entries[i]);
+			staged->entries[i] = NULL;
+		}
+		nb_entries += i;
+		slapi_ch_free_string(&staged->map_group);
+		slapi_ch_free_string(&staged->map_set);
+		slapi_ch_free_string(&staged->name);
+		slapi_ch_free_string(&staged->container_sdn);
+		next = staged->next;
+		free(staged);
+		staged = next;
+	}
+	slapi_log_error(SLAPI_LOG_PLUGIN,
+			cbdata->state->plugin_desc->spd_id,
+			"backend_search_free_staged freed %d stages with %d entries\n",
+			nb_stages,
+			nb_entries);
+	cbdata->staged = NULL;
+}
 static int
 backend_search_cb(Slapi_PBlock *pb)
 {
@@ -1851,6 +1892,7 @@ backend_search_cb(Slapi_PBlock *pb)
 			map_unlock();
 			/* Return newly acquired entries */
 			backend_send_mapped_entries(&cbdata);
+			backend_search_free_staged(&cbdata);
 		} else {
 			slapi_log_error(SLAPI_LOG_PLUGIN,
 					cbdata.state->plugin_desc->spd_id,
@@ -2303,6 +2345,200 @@ backend_betxn_pre_write_cb(Slapi_PBlock *pb)
 	slapi_pblock_get(pb, SLAPI_PLUGIN_PRIVATE, &state);
 	return state->use_be_txns ? backend_write_cb(pb, state) : 0;
 }
+static PRBool
+backend_be_write_ignore(Slapi_PBlock *pb)
+{
+    char *dn = NULL;
+    Slapi_DN *target_sdn = NULL;
+    int check;
+    int i = 0;
+    PRBool ignore = PR_FALSE;
+
+    /* Check if the target DN is a subordinates of
+     * on of the ignored containers
+     */
+    slapi_pblock_get(pb, SLAPI_ORIGINAL_TARGET, &dn);
+    target_sdn = slapi_sdn_new_dn_byval(dn);
+
+    for (i = 0; ignored_containers_sdn[i]; i++) {
+        check  = slapi_sdn_issuffix(target_sdn, ignored_containers_sdn[i]);
+        if (check != 0) {
+            /* This entry is an ignored container */
+            ignore = PR_TRUE;
+            break;
+        }
+    }
+    slapi_sdn_free(&target_sdn);
+    return ignore;
+}
+
+
+static int
+backend_be_pre_write_cb(Slapi_PBlock *pb) {
+    int ret = 0;
+    int lock_status;
+    int lock_count;
+    struct plugin_state *state;
+
+    slapi_pblock_get(pb, SLAPI_PLUGIN_PRIVATE, &state);
+
+    if (wrap_get_call_level() > 0) {
+        return 0;
+    }
+    if (state->ready_to_serve == 0) {
+        /* No data to serve yet */
+        return 0;
+    }
+    if (rw_monitor_enabled() == MAP_MONITOR_DISABLED) {
+        return 0;
+    }
+    if (backend_be_write_ignore(pb)) {
+#if DEBUG_MAP_LOCK
+        slapi_log_error(SLAPI_LOG_FATAL, "schemacompat",
+                "backend_be_pre_write_cb: (%p) operation is not impacting schema compat\n", PR_MyThreadId(), 1);
+#endif
+        return 0;
+    }
+
+
+
+    wrap_inc_call_level();
+
+    lock_status = get_plugin_monitor_status();
+    lock_count = get_plugin_monitor_count();
+    if (lock_status == MAP_RWLOCK_UNINIT) {
+        return 0;
+    }
+
+    if (lock_status == MAP_RWLOCK_FREE) {
+        /* This thread does not hold the map lock */
+        set_plugin_monitor_count(1);
+
+        if (plugin_wrlock() == 0) {
+            ret = 0;
+        } else {
+            slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+                    "backend_be_pre_write_cb: unable to acquire write lock\n");
+            ret = -1;
+        }
+#if DEBUG_MAP_LOCK
+        slapi_log_error(SLAPI_LOG_FATAL, "schemacompat",
+                "backend_be_pre_write_cb: (%p) MAP_RWLOCK_FREE -> MAP_WLOCK_HELD  : count=%d\n", (void *) PR_MyThreadId(), 1);
+#endif
+    } else {
+        set_plugin_monitor_count(lock_count + 1);
+#if DEBUG_MAP_LOCK
+        slapi_log_error(SLAPI_LOG_FATAL, "schema-compat",
+                        "backend_be_pre_write_cb: (%p)  %s --> MAP_WLOCK_HELD  : count=%d\n",
+                (void *) PR_MyThreadId(),
+                (lock_status == MAP_WLOCK_HELD) ? "MAP_WLOCK_HELD": "MAP_RLOCK_HELD",
+                lock_count + 1);
+#endif
+        ret = 0;
+        if (lock_status == MAP_RLOCK_HELD) {
+            /* lock is already acquired in read */
+
+#if DEBUG_MAP_LOCK
+            slapi_log_error(SLAPI_LOG_FATAL, "schema-compat",
+				"map backend_be_pre_write_cb: weird situation map lock is held in read and now required in write mode\n");
+#endif
+            /* First free the lock held in read */
+            ret = plugin_unlock();
+            if (ret) {
+                    slapi_log_error(SLAPI_LOG_FATAL, "schema-compat",
+                            "backend_be_pre_write_cb: fail to unlock plugin lock (%d)\n", ret);
+            }
+
+            /* Second acquire it in write */
+            ret = plugin_wrlock();
+            if (ret) {
+                    slapi_log_error(SLAPI_LOG_FATAL, "schema-compat",
+                            "backend_be_pre_write_cb: fail to write lock plugin lock (%d)\n", ret);
+            }
+        }
+    }
+
+    set_plugin_monitor_status(MAP_WLOCK_HELD);
+    wrap_dec_call_level();
+
+    return ret;
+}
+
+static int
+backend_be_post_write_cb(Slapi_PBlock *pb)
+{
+    int ret = 0;
+    int lock_status;
+    int lock_count;
+    struct plugin_state *state;
+
+    slapi_pblock_get(pb, SLAPI_PLUGIN_PRIVATE, &state);
+
+	if (wrap_get_call_level() > 0) {
+		return 0;
+	}
+	if (state->ready_to_serve == 0) {
+		/* No data to serve yet */
+		return 0;
+	}
+    if (rw_monitor_enabled() == MAP_MONITOR_DISABLED) {
+        return 0;
+    }
+    if (backend_be_write_ignore(pb)) {
+#if DEBUG_MAP_LOCK
+        slapi_log_error(SLAPI_LOG_FATAL, "schemacompat",
+                "backend_be_post_write_cb: (%p) operation was not impacting schema compat\n", PR_MyThreadId(), 1);
+#endif
+        return 0;
+    }
+
+	wrap_inc_call_level();
+
+    lock_status = get_plugin_monitor_status();
+    lock_count = get_plugin_monitor_count();
+    if (lock_status == MAP_RWLOCK_UNINIT) {
+        return 0;
+    }
+
+    if (lock_count == 1) {
+        set_plugin_monitor_status(MAP_RWLOCK_FREE);
+        if (plugin_unlock() == 0) {
+            ret = 0;
+        } else {
+            slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+                    "backend_be_post_write_cb: unable to release write lock\n");
+            ret = -1;
+        }
+#if DEBUG_MAP_LOCK
+        slapi_log_error(SLAPI_LOG_FATAL, "schema-compat",
+                "backend_be_post_write_cb: (%p)  %s --> MAP_RWLOCK_FREE  : count=%d\n",
+                (void *) PR_MyThreadId(),
+                (lock_status == MAP_WLOCK_HELD) ? "MAP_WLOCK_HELD" : (lock_status == MAP_RLOCK_HELD) ? "MAP_RLOCK_HELD" : "MAP_RWLOCK_FREE",
+                0);
+#endif
+    }
+    if (lock_count >= 1) {
+        set_plugin_monitor_count(lock_count - 1);
+#if DEBUG_MAP_LOCK
+        if (lock_count > 1) {
+            slapi_log_error(SLAPI_LOG_FATAL, "schema-compat",
+                    "backend_be_post_write_cb: (%p)  keep %s : count=%d\n",
+                    (void *) PR_MyThreadId(),
+                    (lock_status == MAP_WLOCK_HELD) ? "MAP_WLOCK_HELD" : (lock_status == MAP_RLOCK_HELD) ? "MAP_RLOCK_HELD" : "MAP_RWLOCK_FREE",
+                    lock_count - 1);
+        } else {
+            slapi_log_error(SLAPI_LOG_FATAL, "schema-compat",
+                    "backend_be_post_write_cb: (%p)  is now %s : count=%d\n",
+                    (void *) PR_MyThreadId(),
+                    "MAP_RWLOCK_FREE",
+                    lock_count - 1);
+        }
+#endif
+    }
+	wrap_dec_call_level();
+
+	return ret;
+}
 
 #ifdef USE_PAM
 static int
@@ -2814,6 +3050,42 @@ backend_init_betxn_preop(Slapi_PBlock *pb, struct plugin_state *state)
 	return 0;
 }
 
+int
+backend_init_be_preop(Slapi_PBlock *pb, struct plugin_state *state)
+{
+	slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+	                "hooking up bet preoperation callbacks\n");
+	/* Intercept write requests and return an insufficient-access error for
+	 * attempts to write to anything we're managing. */
+	if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_PRE_ADD_FN,
+			     backend_be_pre_write_cb) != 0) {
+		slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+				"error hooking up betxn pre add callback\n");
+		return -1;
+	}
+	if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_PRE_MODIFY_FN,
+			     backend_be_pre_write_cb) != 0) {
+		slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+				"error hooking up betxn pre modify callback\n");
+		return -1;
+	}
+	if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_PRE_MODRDN_FN,
+			     backend_be_pre_write_cb) != 0) {
+		slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+				"error hooking up betxn pre modrdn callback\n");
+		return -1;
+	}
+	if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_PRE_DELETE_FN,
+			     backend_be_pre_write_cb) != 0) {
+		slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+				"error hooking up betxn pre delete callback\n");
+		return -1;
+	}
+	/* We don't hook abandonment requests. */
+	/* We don't hook unbind requests. */
+	return 0;
+}
+
 int
 backend_init_betxn_postop(Slapi_PBlock *pb, struct plugin_state *state)
 {
@@ -2821,6 +3093,40 @@ backend_init_betxn_postop(Slapi_PBlock *pb, struct plugin_state *state)
 			"hooking up betxn postoperation callbacks\n");
 	return backend_shr_betxn_postop_init(pb, state);
 }
+
+
+int
+backend_init_be_postop(Slapi_PBlock *pb, struct plugin_state *state)
+{
+	if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_POST_ADD_FN,
+			     backend_be_post_write_cb) != 0) {
+		slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+				"error hooking up betxn post add callback\n");
+		return -1;
+	}
+	if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_POST_MODIFY_FN,
+			     backend_be_post_write_cb) != 0) {
+		slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+				"error hooking up betxn post modify "
+				"callback\n");
+		return -1;
+	}
+	if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_POST_MODRDN_FN,
+			     backend_be_post_write_cb) != 0) {
+		slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+				"error hooking up betxn post modrdn "
+				"callback\n");
+		return -1;
+	}
+	if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_POST_DELETE_FN,
+			     backend_be_post_write_cb) != 0) {
+		slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+				"error hooking up betxn post delete "
+				"callback\n");
+		return -1;
+	}
+	return 0;
+}
 #endif
 
 int
