@@ -37,6 +37,9 @@
 #endif
 
 #include <rpc/xdr.h>
+#include <prtypes.h>
+/* NSPR private API for thread-local variables */
+#include <private/prpriv.h>
 #include "../yp/yp.h"
 
 #ifdef HAVE_TCPD_H
@@ -777,8 +780,6 @@ static void
 backend_shr_data_initialize_thread(time_t when, void *arg)
 {
 	struct backend_shr_data_init_cbdata *cbdata = (struct backend_shr_data_init_cbdata *)arg;
-	PRThread *thread = NULL;
-
 	if (slapi_is_shutting_down()) {
 		return;
 	}
@@ -2866,3 +2867,201 @@ backend_shr_internal_postop_init(Slapi_PBlock *pb, struct plugin_state *state)
 	}
 	return 0;
 }
+
+int
+PR_MyThreadId(void)
+{
+    PRThread *thr = PR_GetCurrentThread();
+    PRUint32 myself = PR_GetThreadID(thr);
+    return myself;
+}
+PRBool use_lock_status = PR_FALSE;
+static unsigned int thread_plugin_lock_status;
+PRBool use_lock_count = PR_FALSE;
+static unsigned int thread_plugin_lock_count;
+/*
+ * https://developer.mozilla.org/en-US/docs/Mozilla/Projects/NSPR/Reference/PR_NewThreadPrivateIndex
+ * It is called each time:
+ *  - PR_SetThreadPrivate is call with a not NULL private value
+ *  - on thread exit
+ */
+static void
+lock_status_free(void *ptr)
+{
+    int *lock_status = ptr;
+    if (lock_status) {
+        slapi_ch_free((void **)&lock_status);
+    }
+}
+static void
+lock_count_free(void *ptr)
+{
+    int *lock_count = ptr;
+    if (lock_count) {
+        slapi_ch_free((void **)&lock_count);
+    }
+}
+
+/* This is used to ignore some write operations
+ * if they target subordinates entry of ignored containers
+ */
+Slapi_DN **ignored_containers_sdn = NULL;
+void
+init_map_lock(void)
+{
+#if DEBUG_MAP_LOCK
+    /* The plugin lock is initialized as free */
+    slapi_log_error(SLAPI_LOG_FATAL, "init_plugin_lock",
+            "thread_id = %p\n", PR_MyThreadId());
+#endif
+    if (ignored_containers_sdn == NULL) {
+        /* allocates ignored_containers_sdn of Slapi_DN* only once
+         * It will be used later in be_pre/post_write_cb
+         */
+        int cnt;
+        char *ignored_containers[4] = { "cn=config", "cn=schema", "cn=changelog", NULL};
+
+        for (cnt = 0; ignored_containers[cnt]; cnt++);
+        ignored_containers_sdn = (Slapi_DN **) slapi_ch_calloc(cnt + 1, sizeof(Slapi_DN *));
+
+        for (cnt = 0; ignored_containers[cnt]; cnt++) {
+            ignored_containers_sdn[cnt] = slapi_sdn_new_dn_byval(ignored_containers[cnt]);
+        }
+    }
+
+    if (PR_NewThreadPrivateIndex(&thread_plugin_lock_status, lock_status_free) == PR_SUCCESS) {
+        use_lock_status = PR_TRUE;
+    }
+    if (PR_NewThreadPrivateIndex(&thread_plugin_lock_count, lock_count_free) == PR_SUCCESS) {
+        use_lock_count = PR_TRUE;
+    }
+
+#if DEBUG_MAP_LOCK
+    slapi_log_error(SLAPI_LOG_FATAL, "init_plugin_lock",
+            "thread_plugin_lock_status = %d\n", thread_plugin_lock_status);
+    slapi_log_error(SLAPI_LOG_FATAL, "init_map_lock",
+            "thread_plugin_lock_count = %d\n", thread_plugin_lock_count);
+#endif
+}
+
+int
+rw_monitor_enabled(void)
+{
+    if (use_lock_status)
+        return (int) MAP_MONITOR_ENABLED;
+    else
+        return (int) MAP_MONITOR_DISABLED;
+}
+
+int
+get_plugin_monitor_status(void)
+{
+    int ret;
+
+    if (use_lock_status) {
+        int *lock_status_p;
+        lock_status_p = (int *) PR_GetThreadPrivate(thread_plugin_lock_status);
+        if (lock_status_p == NULL) {
+            /* if it was not initialized then allocates a structure
+             * that will be private to that thread.
+             * Later when the structure is retrieved (PR_GetThreadPrivate), the
+             * content of the structure can be read/write without allocating
+             * this private structure PR_SetThreadPrivate.
+             * So this branch is processed one time per each thread
+             */
+            lock_status_p = (int *) slapi_ch_calloc(1, sizeof (int));
+            PR_SetThreadPrivate(thread_plugin_lock_status, (void *) lock_status_p);
+            *lock_status_p = MAP_RWLOCK_FREE;
+        }
+        ret = *lock_status_p;
+    } else
+        ret = (int) MAP_RWLOCK_UNINIT;
+#if DEBUG_MAP_LOCK
+    slapi_log_error(SLAPI_LOG_FATAL, "get_plugin_monitor_status",
+            "(%s) lock_status = %d (%p)\n", use_lock_status ? "TRUE" : "FALSE", ret, (void *) PR_MyThreadId());
+#endif
+    return ret;
+}
+
+void
+set_plugin_monitor_status(int lock_status)
+{
+#if DEBUG_MAP_LOCK
+    slapi_log_error(SLAPI_LOG_FATAL, "set_plugin_monitor_status",
+            "(%s) lock_status = %d --> %d (%p)\n", use_lock_status ? "TRUE" : "FALSE", get_plugin_monitor_status(), (int) lock_status, (void *) PR_MyThreadId());
+#endif
+
+    if (use_lock_status) {
+        int *lock_status_p;
+
+        lock_status_p = (int *) PR_GetThreadPrivate(thread_plugin_lock_status);
+        if (lock_status_p == NULL) {
+            /* if it was not initialized then allocates a structure
+             * that will be private to that thread.
+             * Later when the structure is retrieved (PR_GetThreadPrivate), the
+             * content of the structure can be read/write without allocating
+             * this private structure PR_SetThreadPrivate.
+             * So this branch is processed one time per each thread
+             */
+            lock_status_p = (int *) slapi_ch_calloc(1, sizeof (int));
+            PR_SetThreadPrivate(thread_plugin_lock_status, (void *) lock_status_p);
+        }
+        *lock_status_p = lock_status;
+    }
+}
+
+int
+get_plugin_monitor_count(void)
+{
+    int ret;
+
+    if (use_lock_count) {
+        int *lock_count;
+        lock_count = (int *) PR_GetThreadPrivate(thread_plugin_lock_count);
+        if (lock_count == NULL) {
+            /* if it was not initialized then allocates a structure
+             * that will be private to that thread.
+             * Later when the structure is retrieved (PR_GetThreadPrivate), the
+             * content of the structure can be read/write without allocating
+             * this private structure PR_SetThreadPrivate.
+             * So this branch is processed one time per each thread
+             */
+            lock_count = (int *) slapi_ch_calloc(1, sizeof (int));
+            PR_SetThreadPrivate(thread_plugin_lock_count, (void *) lock_count);
+            *lock_count = 0;
+        }
+        ret = *lock_count;
+    } else
+        ret = (int) MAP_RWLOCK_UNINIT;
+#if DEBUG_MAP_LOCK
+    slapi_log_error(SLAPI_LOG_FATAL, "get_plugin_monitor_count",
+            "(%s) lock_count = %d (%p)\n", use_lock_count ? "TRUE" : "FALSE", ret, (void *) PR_MyThreadId());
+#endif
+    return ret;
+}
+
+void
+set_plugin_monitor_count(int lock_count)
+{
+#if DEBUG_MAP_LOCK
+    slapi_log_error(SLAPI_LOG_FATAL, "set_plugin_monitor_count",
+            "(%s) lock_count = %d --> %d (%p)\n", use_lock_count ? "TRUE" : "FALSE", get_plugin_monitor_count(), (int) lock_count, (void *) PR_MyThreadId());
+#endif
+    if (use_lock_count) {
+        int *lock_count_p;
+
+        lock_count_p = (int *) PR_GetThreadPrivate(thread_plugin_lock_count);
+        if (lock_count_p == NULL) {
+            /* if it was not initialized then allocates a structure
+             * that will be private to that thread.
+             * Later when the structure is retrieved (PR_GetThreadPrivate), the
+             * content of the structure can be read/write without allocating
+             * this private structure PR_SetThreadPrivate.
+             * So this branch is processed one time per each thread
+             */
+            lock_count_p = (int *) slapi_ch_calloc(1, sizeof (int));
+            PR_SetThreadPrivate(thread_plugin_lock_count, (void *) lock_count_p);
+        }
+        *lock_count_p = lock_count;
+    }
+}
