@@ -34,6 +34,8 @@
 #include <rpc/rpc.h>
 #include "../../yp/yp.h"
 
+typedef int (*xdrrec_proc_t)(void *, void *, int);
+
 static struct sockaddr_in server;
 static int connected;
 
@@ -162,8 +164,7 @@ writejunk(char *fd, char *data, int size)
 static int
 all(CLIENT *client, FILE *output, int argc, char **argv)
 {
-	ypresp_all *ret;
-	int i, sock;
+	int sock;
 	ypreq_nokey nokey;
 	struct rpc_msg req, rep;
 	struct ypresp_all resp;
@@ -189,7 +190,7 @@ all(CLIENT *client, FILE *output, int argc, char **argv)
 	}
 
 	memset(&s, 0, sizeof(s));
-	xdrrec_create(&s, 0, 0, (char *) &sock, &readjunk, &writejunk);
+	xdrrec_create(&s, 0, 0, (char *) &sock, (xdrrec_proc_t) &readjunk, (xdrrec_proc_t) &writejunk);
 	s.x_op = XDR_ENCODE;
 
 	memset(&req, 0, sizeof(req));
@@ -218,7 +219,7 @@ all(CLIENT *client, FILE *output, int argc, char **argv)
 	xdr_destroy(&s);
 
 	memset(&s, 0, sizeof(s));
-	xdrrec_create(&s, 0, 0, (char *) &sock, &readjunk, &writejunk);
+	xdrrec_create(&s, 0, 0, (char *) &sock, (xdrrec_proc_t) &readjunk, (xdrrec_proc_t) &writejunk);
 	s.x_op = XDR_DECODE;
 	xdrrec_skiprecord(&s);
 
